
require 'torch'
require 'nn'

local DAG, parent = torch.class('nn.DAG', 'nn.Container')

function DAG:__init()
   parent.__init(self)
   self.pred = {}
   self.succ = {}
end

function DAG:addEdge(a, b)
   self.sorted = nil
   local pred, succ = self.pred, self.succ
   if not pred[a] and not succ[a] then
      self:add(a)
   end
   if not pred[b] and not succ[b] then
      self:add(b)
   end
   pred[b] = pred[b] or {}
   pred[b][#pred[b] + 1] = a
   succ[a] = succ[a] or {}
   succ[a][#succ[a] + 1] = b
end

-- Apply f on t recursively; use the corresponding a1 and a2 elements
-- (i.e. same keys) as second and third parameters to f when
-- available; return the results from f, organized in a similarly
-- nested table.
function DAG:applyOnModules(f, t, a1, a2)
   if torch.type(t) == 'table' then
      local result = {}
      for k, s in pairs(t) do
         result[k] = self:applyOnModules(f, s, a1 and a1[k], a2 and a2[k])
      end
      return result
   else
      return f(t, a1, a2)
   end
end

function DAG:setInput(i)
   self.sorted = nil
   self.inputModules = i
   self:applyOnModules(
      function(m)
         if not self.succ[m] or #self.succ[m] == 0 then
            error('Input modules must have outgoing  edges.')
         end
         if self.pred[m] and #self.pred[m] > 0 then
            error('Input modules cannog have incoming edges.')
         end
      end,
      self.inputModules
   )
end

function DAG:setOutput(o)
   self.sorted = nil
   self.outputModules = o
   self:applyOnModules(
      function(m)
         if not self.pred[m] or #self.pred[m] == 0 then
            error('Output module must have incoming edges.')
         end
         if self.succ[m] and #self.succ[m] > 0 then
            error('Output module cannot have outgoing edges.')
         end
      end,
      self.outputModules
   )
end

function DAG:sort()
   if self.sorted then
      return
   end

   local distance = {}

   self:applyOnModules(function(m) distance[m] = 1 end, self.inputModules)

   local nc

   repeat
      nc = 0
      for i, isucc in pairs(self.succ) do
         for _, j in pairs(isucc) do
            if distance[i] and (not distance[j] or distance[j] < distance[i] + 1) then
               distance[j] = distance[i] + 1
               nc = nc + 1
            end
         end
      end
   until nc == 0

   self.sorted = { }
   for i, d in pairs(distance) do
      table.insert(self.sorted, { d, i })
   end

   table.sort(self.sorted, function(a, b) return a[1] < b[1] end)
   for i, a in ipairs(self.sorted) do self.sorted[i] = a[2] end
end

function DAG:print()
   self:sort()

   for i, d in ipairs(self.sorted) do
      print('#' .. i .. ' -> ' .. torch.type(d))
   end
end

function DAG:updateOutput(input)
   self:sort()

   self:applyOnModules(function(m, i) m:updateOutput(i) end, self.inputModules, input)

   for _, d in ipairs(self.sorted) do
      if self.pred[d] then
         if #self.pred[d] == 1 then
            d:updateOutput(self.pred[d][1].output)
         elseif #self.pred[d] > 1 then
            local c = {}
            for k = 1, #self.pred[d] do
               c[k] = self.pred[d][k].output
            end
            d:updateOutput(c)
         end
      end
   end

   self.output = self:applyOnModules(function(m) return m.output end, self.outputModules)

   return self.output
end

function DAG:updateGradInput(input, gradOutput)
   self:sort()

   self:applyOnModules(function(m, i, go) m:updateGradInput(i, go) end, self.outputModules, input, gradOutput)

   for k = self.sorted, 1, -1 do
      local m = sorted[k]
      if self.succ[d] then
         if #self.succ[d] == 1 then
            d:updateGradInput(self.succ[d][1].gradInput)
         elseif #self.succ[d] > 1 then
            local sum
            for k = 1, #self.succ[d] do
               if sum then
                  sum:add(self.succ[d][k].gradInput)
               else
                  sum = self.succ[d][k].gradInput:clone()
               end
            end
            d:updateGradInput(sum)
         end
      end
   end

   self.gradInput = self:applyOnModules(function(m) return m.gradInput end, self.inputModules)

   return self.gradInput
end

return DAG
